/* Baduki - Computer Baduk(Go) Program 
   Copyright(c) 1998-1999 Lim Jaebum <artist@soback.kornet21.net>

   This program is free software; you can redistribute it and'or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
   USA */

/* $Id: dialog.c,v 1.10 1999/04/12 16:31:47 artist Exp artist $ */ 


#include <config.h>
#include "intldefs.h"

#include <stdio.h>
#include <gtk/gtk.h>
#include "baduki.h"
#include "gpointer.h"

void bdk_set_handicap(GtkWidget *widget, gpointer data);
void bdk_set_playlevel(GtkWidget *widget, gpointer data);
void bdk_set_black_type(GtkWidget *widget, gpointer data);
void bdk_set_white_type(GtkWidget *widget, gpointer data);
extern int player_type[];
extern int new_handicap, showTrees, show_altmoves, showMoveReason, playLevel;
extern char player_name[][80];

#define OPT_SHOW_THINK 0
#define OPT_SHOW_ALTMOVES 1
#define OPT_SHOW_REASON 2

void set_options(GtkCheckButton *check, gpointer data)
{
  switch (GPOINTER_TO_INT(data))
    {
    case OPT_SHOW_THINK:
      showTrees = !showTrees;
      break;
    case OPT_SHOW_ALTMOVES:
      show_altmoves = !show_altmoves;
      break;
    case OPT_SHOW_REASON:
      showMoveReason = !showMoveReason;
      break;
    }
}


static GtkWidget* create_handicap_menu (void)
{
  GtkWidget *menu;
  GtkWidget *menuitem;
  GSList *group;
  char buf[32];
  int i;

  menu = gtk_menu_new ();
  group = NULL;

  for (i = 0; i < 10; i++) {
    if (i == 0)
      sprintf(buf, _("Handicap: 0 + 5.5komi"));
    else if (i == 1)
      sprintf(buf, _("Handicap: 1 + 0.5komi"));
    else 
      sprintf(buf, _("Handicap: %d + 0.5komi"), i);
      
    menuitem = gtk_radio_menu_item_new_with_label (group, buf);
    gtk_signal_connect(GTK_OBJECT (menuitem), "activate",
	GTK_SIGNAL_FUNC (bdk_set_handicap), GINT_TO_POINTER(i));
    group = gtk_radio_menu_item_group (GTK_RADIO_MENU_ITEM (menuitem));
    gtk_menu_append (GTK_MENU (menu), menuitem);
    gtk_widget_show (menuitem);
  }

  return menu;
}

static GtkWidget* create_playlevel_menu (void)
{
  GtkWidget *menu;
  GtkWidget *menuitem;
  GSList *group;
  char buf[32];
  int i;

  menu = gtk_menu_new ();
  group = NULL;

  for (i = 1; i <= 12; i++) {
    sprintf(buf, _("Level : %d"), i);
      
    menuitem = gtk_radio_menu_item_new_with_label (group, buf);
    gtk_signal_connect(GTK_OBJECT (menuitem), "activate",
	GTK_SIGNAL_FUNC (bdk_set_playlevel), GINT_TO_POINTER(i));
    group = gtk_radio_menu_item_group (GTK_RADIO_MENU_ITEM (menuitem));
    gtk_menu_append (GTK_MENU (menu), menuitem);
    gtk_widget_show (menuitem);
  }

  return menu;
}

const char *ptype_name[] = {N_("Baduki"), N_("Human"), N_("GMP"), N_("IGS")};

static GtkWidget* create_player_type_menu(int color)
{
  GtkWidget *menu;
  GtkWidget *menuitem;
  GSList *group;
  int i;

  menu = gtk_menu_new ();
  group = NULL;

  for (i = 0; i < 4; i++) {
    menuitem = gtk_radio_menu_item_new_with_label (group, _(ptype_name[i]));
    if (color == BLACK)
      gtk_signal_connect(GTK_OBJECT (menuitem), "activate",
	GTK_SIGNAL_FUNC (bdk_set_black_type), GINT_TO_POINTER(i));
    else
      gtk_signal_connect(GTK_OBJECT (menuitem), "activate",
	GTK_SIGNAL_FUNC (bdk_set_white_type), GINT_TO_POINTER(i));
    group = gtk_radio_menu_item_group (GTK_RADIO_MENU_ITEM (menuitem));
    gtk_menu_append (GTK_MENU (menu), menuitem);
    gtk_widget_show (menuitem);
  }

  return menu;
}

void set_player_name(GtkEntry *entry, gpointer data)
{
  int p = GPOINTER_TO_INT(data);
  
  if (p != BLACK && p != WHITE)
    return;
  
  sprintf(player_name[p], gtk_entry_get_text(entry));
}

void option_dialog()
{
  GtkWidget *dialog;
  GtkWidget *dialog_vbox1;
  GtkWidget *vbox1;
  GtkWidget *hbox1;
  GtkWidget *frame1;
  GtkWidget *vbox2;
  GtkWidget *optionmenu1;
  GtkWidget *entry1;
  GtkWidget *frame2;
  GtkWidget *vbox3;
  GtkWidget *optionmenu2;
  GtkWidget *entry2;
  GtkWidget *optionmenu3;
  GtkWidget *checkbutton1;
  GtkWidget *checkbutton2;
  GtkWidget *checkbutton3;
  GtkWidget *dialog_action_area1;
  GtkWidget *button;
  GtkTooltips *tooltips;

  tooltips = gtk_tooltips_new();

  dialog = gtk_dialog_new();
  gtk_object_set_data(GTK_OBJECT(dialog), "Option", dialog);
  gtk_window_set_title(GTK_WINDOW(dialog), "Option");

  dialog_vbox1 = GTK_DIALOG (dialog)->vbox;
  gtk_object_set_data(GTK_OBJECT(dialog), "dialog_vbox1", dialog_vbox1);
  gtk_widget_show(dialog_vbox1);

  vbox1 = gtk_vbox_new(FALSE, 0);
  gtk_object_set_data(GTK_OBJECT(dialog), "vbox1", vbox1);
  gtk_widget_show(vbox1);
  gtk_box_pack_start(GTK_BOX(dialog_vbox1), vbox1, TRUE, TRUE, 0);

  hbox1 = gtk_hbox_new(FALSE, 0);
  gtk_object_set_data(GTK_OBJECT(dialog), "hbox1", hbox1);
  gtk_widget_show(hbox1);
  gtk_box_pack_start(GTK_BOX(vbox1), hbox1, TRUE, TRUE, 0);

  frame1 = gtk_frame_new(_("Black"));
  gtk_frame_set_label_align(GTK_FRAME(frame1), 0.5, 0.5);
  gtk_object_set_data(GTK_OBJECT(dialog), "frame1", frame1);
  gtk_widget_show(frame1);
  gtk_box_pack_start(GTK_BOX(hbox1), frame1, TRUE, TRUE, 0);

  vbox2 = gtk_vbox_new(FALSE, 0);
  gtk_object_set_data(GTK_OBJECT(dialog), "vbox2", vbox2);
  gtk_widget_show(vbox2);
  gtk_container_add(GTK_CONTAINER(frame1), vbox2);

  optionmenu1 = gtk_option_menu_new();
  gtk_option_menu_set_menu (GTK_OPTION_MENU (optionmenu1),
	create_player_type_menu(BLACK));
  gtk_option_menu_set_history (GTK_OPTION_MENU (optionmenu1), player_type[BLACK]);
  gtk_box_pack_start(GTK_BOX(vbox2), optionmenu1, TRUE, TRUE, 0);
  gtk_widget_show (optionmenu1);
  gtk_tooltips_set_tip(tooltips, optionmenu1, _("Player Type (Black)"), NULL);

  entry1 = gtk_entry_new();
  if (player_name[BLACK] && *player_name[BLACK])
    gtk_entry_set_text(GTK_ENTRY(entry1), player_name[BLACK]);
  gtk_object_set_data(GTK_OBJECT(dialog), "entry1", entry1);
  gtk_signal_connect (GTK_OBJECT(entry1), "changed",
	GTK_SIGNAL_FUNC(set_player_name), (gpointer) BLACK);
  gtk_widget_show(entry1);
  gtk_box_pack_start(GTK_BOX(vbox2), entry1, TRUE, TRUE, 0);
  gtk_tooltips_set_tip(tooltips, entry1, _("Black Player's Name"), NULL);

  frame2 = gtk_frame_new(_("White"));
  gtk_frame_set_label_align(GTK_FRAME(frame2), 0.5, 0.5);
  gtk_object_set_data(GTK_OBJECT(dialog), "frame2", frame2);
  gtk_widget_show(frame2);
  gtk_box_pack_start(GTK_BOX(hbox1), frame2, TRUE, TRUE, 0);

  vbox3 = gtk_vbox_new(FALSE, 0);
  gtk_object_set_data(GTK_OBJECT(dialog), "vbox3", vbox3);
  gtk_widget_show(vbox3);
  gtk_container_add(GTK_CONTAINER(frame2), vbox3);

  optionmenu2 = gtk_option_menu_new();
  gtk_option_menu_set_menu (GTK_OPTION_MENU (optionmenu2),
	create_player_type_menu(WHITE));
  gtk_option_menu_set_history (GTK_OPTION_MENU (optionmenu2), player_type[WHITE]);
  gtk_box_pack_start(GTK_BOX(vbox3), optionmenu2, TRUE, TRUE, 0);
  gtk_widget_show (optionmenu2);
  gtk_tooltips_set_tip(tooltips, optionmenu2, _("Player Type (White)"), NULL);

  entry2 = gtk_entry_new();
  if (player_name[WHITE] && *player_name[WHITE])
    gtk_entry_set_text(GTK_ENTRY(entry2), player_name[WHITE]);
  gtk_object_set_data(GTK_OBJECT(dialog), "entry2", entry2);
  gtk_signal_connect (GTK_OBJECT(entry2), "changed",
	GTK_SIGNAL_FUNC(set_player_name), (gpointer) WHITE);
  gtk_widget_show(entry2);
  gtk_box_pack_start(GTK_BOX(vbox3), entry2, TRUE, TRUE, 0);
  gtk_tooltips_set_tip(tooltips, entry2, _("White Player's Name"), NULL);

  optionmenu3 = gtk_option_menu_new();
  gtk_option_menu_set_menu (GTK_OPTION_MENU (optionmenu3),
	create_handicap_menu ());
  gtk_option_menu_set_history (GTK_OPTION_MENU (optionmenu3), new_handicap);
  gtk_box_pack_start(GTK_BOX(vbox1), optionmenu3, TRUE, TRUE, 0);
  gtk_widget_show (optionmenu3);
  
  optionmenu3 = gtk_option_menu_new();
  gtk_option_menu_set_menu (GTK_OPTION_MENU (optionmenu3),
	create_playlevel_menu ());
  gtk_option_menu_set_history (GTK_OPTION_MENU (optionmenu3), playLevel - 1);
  gtk_box_pack_start(GTK_BOX(vbox1), optionmenu3, TRUE, TRUE, 0);
  gtk_widget_show (optionmenu3);
  
  checkbutton1 = gtk_check_button_new_with_label(_("Show Think"));
  gtk_object_set_data(GTK_OBJECT(dialog), "checkbutton1", checkbutton1);
  gtk_widget_show(checkbutton1);
  gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (checkbutton1),
                               showTrees);
  gtk_signal_connect (GTK_OBJECT (checkbutton1), "toggled",
                      GTK_SIGNAL_FUNC (set_options),
                      (gpointer) OPT_SHOW_THINK);
  gtk_box_pack_start(GTK_BOX(vbox1), checkbutton1, TRUE, TRUE, 0);

  checkbutton2 = gtk_check_button_new_with_label(_("Show alt. moves."));
  gtk_object_set_data(GTK_OBJECT(dialog), "checkbutton2", checkbutton2);
  gtk_widget_show(checkbutton2);
  gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (checkbutton2),
                               show_altmoves);
  gtk_signal_connect (GTK_OBJECT (checkbutton2), "toggled",
                      GTK_SIGNAL_FUNC (set_options),
                      (gpointer) OPT_SHOW_ALTMOVES);
  gtk_box_pack_start(GTK_BOX(vbox1), checkbutton2, TRUE, TRUE, 0);

  checkbutton3 = gtk_check_button_new_with_label(_("Show Reason"));
  gtk_object_set_data(GTK_OBJECT(dialog), "checkbutton3", checkbutton3);
  gtk_widget_show(checkbutton3);
  gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (checkbutton3),
                               showMoveReason);
  gtk_signal_connect (GTK_OBJECT (checkbutton3), "toggled",
                      GTK_SIGNAL_FUNC (set_options),
                      (gpointer) OPT_SHOW_REASON);
  gtk_box_pack_start(GTK_BOX(vbox1), checkbutton3, TRUE, TRUE, 0);

  dialog_action_area1 = GTK_DIALOG (dialog)->action_area;
  gtk_object_set_data(GTK_OBJECT(dialog), "dialog_action_area1", dialog_action_area1);
  gtk_widget_show(dialog_action_area1);
  gtk_container_border_width(GTK_CONTAINER(dialog_action_area1), 10);

  button = gtk_button_new_with_label(_("OK"));
  gtk_widget_set_usize(button, 80, -1);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), button, FALSE,
                     FALSE, 14);
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_widget_grab_default (button);
  gtk_widget_show(button);

  gtk_signal_connect_object (GTK_OBJECT (dialog), "delete_event",
                             GTK_SIGNAL_FUNC(gtk_widget_destroy),
                             GTK_OBJECT (dialog));
  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
                             GTK_SIGNAL_FUNC(gtk_widget_destroy),
                             GTK_OBJECT (dialog));
  gtk_widget_show(dialog);
}

