/* Baduki - Computer Baduk(Go) Program 
   Copyright(c) 1998-1999 Lim Jaebum <artist@soback.kornet21.net> */

/* $Id: gtkgoboard.c,v 1.8 1999/02/06 12:32:42 artist Exp $ */

/* gtkgoboard.c -- GTK+ go board widget.
   Copyright (C) 1998 Changwoo Ryu

   src/drawStone.c, part of Complete Goban (game program)
   Copyright (C) 1995-1996 William Shubert.

   This program is free software; you can redistribute it and'or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
   USA */

/* Written by Changwoo Ryu <cwryu@adam.kaist.ac.kr>. */

/* Stone drawing is derived from William Shubert's CGoban. */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */

#include "gtkgoboard.h"

#include <gtk/gtksignal.h>
#include <math.h>
#include <string.h>
#include <stdio.h>

enum
{
  PRESSED,
  RELEASED,
  CLICKED,
  ENTER,
  LEAVE,
  LAST_SIGNAL
};


#define DEFAULT_CELL_SIZE	18


#define HORIZONTAL_RIGHT_LINE	(1 << 0)
#define HORIZONTAL_LEFT_LINE	(1 << 1)
#define HORIZONTAL_FULL_LINE	(1 << 2)
#define HORIZONTAL_LINE		(HORIZONTAL_RIGHT_LINE | HORIZONTAL_LEFT_LINE | HORIZONTAL_FULL_LINE)
#define VERTICAL_LOWER_LINE	(1 << 3)
#define VERTICAL_UPPER_LINE	(1 << 4)
#define VERTICAL_FULL_LINE	(1 << 5)
#define VERTICAL_LINE		(VERTICAL_LOWER_LINE | VERTICAL_UPPER_LINE | VERTICAL_FULL_LINE)
#define HANDICAP_POINT		(1 << 6)

#define NO_LINE_MARKUP_MASK	(GTK_GO_BOARD_MA_MASK \
				 | GTK_GO_BOARD_TR_MASK \
				 | GTK_GO_BOARD_CR_MASK \
				 | GTK_GO_BOARD_SQ_MASK \
				 | GTK_GO_BOARD_SL_MASK)
#define IS_NO_LINE_MARKUP(m)	(m & NO_LINE_MARKUP_MASK)

/* Initialize...  */
static void gtk_go_board_class_init     (GtkGoBoardClass	*klass);
static void gtk_go_board_init           (GtkGoBoard     	*go_board);

/* GtkObject signals...  */
static void gtk_go_board_destroy	(GtkObject *object);
static void gtk_go_board_finalize 	(GtkObject * object);

/* GtkWidget signals...  */
static gint gtk_go_board_button_press	(GtkWidget        	*widget,
					 GdkEventButton   	*event);
static gint gtk_go_board_button_release	(GtkWidget        	*widget,
					 GdkEventButton   	*event);
static gint gtk_go_board_motion_notify	(GtkWidget        	*widget,
					 GdkEventMotion   	*event);
static gint gtk_go_board_leave_notify	(GtkWidget        	*widget,
					 GdkEventCrossing       *event);

static gint gtk_go_board_expose         (GtkWidget		*widget,
					 GdkEventExpose		*event);
static void gtk_go_board_size_request   (GtkWidget    		*widget,
					 GtkRequisition		*requisition);
static void gtk_go_board_size_allocate  (GtkWidget    		*widget,
					 GtkAllocation		*allocation);
static void gtk_go_board_realize	(GtkWidget    		*widget);

/* Signal marshallers...  */
typedef void (*GtkGoBoardSignal1) (GtkObject *object,
				   gint x, gint y, gpointer data);
typedef void (*GtkGoBoardSignal2) (GtkObject *object,
				   gint x, gint y, gint bt, gpointer data);
static void gtk_go_board_marshaller_1	(GtkObject *object,
					 GtkSignalFunc func,
					 gpointer func_data,
					 GtkArg *args);
static void gtk_go_board_marshaller_2	(GtkObject *object,
					 GtkSignalFunc func,
					 gpointer func_data,
					 GtkArg *args);

static void gtk_go_board_init_cells 	(GtkGoBoard *go_board);
static void gtk_go_board_paint 		(GtkWidget    *widget,
					 GdkRectangle *area);

static gboolean gtk_go_board_greys_alloc	(GtkGoBoard *go_board);
static void gtk_go_board_greys_free		(GtkGoBoard *go_board);

static void gtk_go_board_update_pixmap 	(GtkGoBoard *go_board);
static float gtk_go_board_calc_reflection (int x, int y, int r,
					   float *lambertian, float *z);
static gint gtk_go_board_max_coord_width (GtkGoBoard *go_board);
static gint gtk_go_board_max_coord_height (GtkGoBoard *go_board);
static gchar *gtk_go_board_row_coord_string (GtkGoBoard *go_board, int n);
static gchar *gtk_go_board_col_coord_string (GtkGoBoard *go_board, int n);

static gboolean gtk_go_board_get_board_coordinate (GtkGoBoard *go_board,
						   int x, int y,
						   int *rx, int *ry);
static void gtk_go_board_update_layout (GtkGoBoard *go_board, gint width, gint height);



static GtkDrawingAreaClass *parent_class;
static void (* parent_size_allocate) (GtkWidget *widget, GtkAllocation *allocation);
static void (* parent_realize) (GtkWidget *widget);

static guint go_board_signals[LAST_SIGNAL] = { 0 };


GtkType
gtk_go_board_get_type ()
{
  static guint go_board_type = 0;

  if (!go_board_type)
    {
      GtkTypeInfo go_board_info =
      {
	"GtkGoBoard",
	sizeof (GtkGoBoard),
	sizeof (GtkGoBoardClass),
	(GtkClassInitFunc) gtk_go_board_class_init,
	(GtkObjectInitFunc) gtk_go_board_init,
	(GtkArgSetFunc) NULL,
	(GtkArgGetFunc) NULL
      };

      go_board_type = gtk_type_unique (gtk_drawing_area_get_type (), &go_board_info);
    }

  return go_board_type;
}


static void
gtk_go_board_class_init (GtkGoBoardClass *class)
{
  GtkObjectClass *object_class;
  GtkWidgetClass *widget_class;
  GtkDrawingAreaClass *drawing_area_class;


  object_class = (GtkObjectClass*) class;
  widget_class = (GtkWidgetClass*) class;
  drawing_area_class = (GtkDrawingAreaClass*) drawing_area_class;

  parent_class = gtk_type_class (gtk_drawing_area_get_type ());
  go_board_signals[PRESSED] =
    gtk_signal_new ("pressed",
		    GTK_RUN_FIRST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GtkGoBoardClass, pressed),
		    gtk_go_board_marshaller_2,
		    GTK_TYPE_NONE, 3,
		    GTK_TYPE_INT,
		    GTK_TYPE_INT,
		    GTK_TYPE_INT);
  go_board_signals[RELEASED] =
    gtk_signal_new ("released",
		    GTK_RUN_FIRST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GtkGoBoardClass, pressed),
		    gtk_go_board_marshaller_2,
		    GTK_TYPE_NONE, 3,
		    GTK_TYPE_INT,
		    GTK_TYPE_INT,
		    GTK_TYPE_INT);
  go_board_signals[CLICKED] =
    gtk_signal_new ("clicked",
		    GTK_RUN_FIRST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GtkGoBoardClass, pressed),
		    gtk_go_board_marshaller_1,
		    GTK_TYPE_NONE, 2,
		    GTK_TYPE_INT,
		    GTK_TYPE_INT);
  go_board_signals[ENTER] =
    gtk_signal_new ("enter",
		    GTK_RUN_FIRST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GtkGoBoardClass, enter),
		    gtk_go_board_marshaller_1,
		    GTK_TYPE_NONE, 2,
		    GTK_TYPE_INT,
		    GTK_TYPE_INT);
  go_board_signals[LEAVE] =
    gtk_signal_new ("leave",
		    GTK_RUN_FIRST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GtkGoBoardClass, leave),
		    gtk_go_board_marshaller_1,
		    GTK_TYPE_NONE, 2,
		    GTK_TYPE_INT,
		    GTK_TYPE_INT);

  gtk_object_class_add_signals (object_class, go_board_signals, LAST_SIGNAL);

  object_class->destroy = gtk_go_board_destroy;
  object_class->finalize = gtk_go_board_finalize;

  widget_class->expose_event = gtk_go_board_expose;
  widget_class->button_press_event = gtk_go_board_button_press;
  widget_class->button_release_event = gtk_go_board_button_release;
  widget_class->motion_notify_event = gtk_go_board_motion_notify;
  widget_class->leave_notify_event = gtk_go_board_leave_notify;

  widget_class->size_request = gtk_go_board_size_request;
  parent_size_allocate = widget_class->size_allocate;
  widget_class->size_allocate = gtk_go_board_size_allocate;
  parent_realize = widget_class->realize;
  widget_class->realize = gtk_go_board_realize;

  class->pressed = NULL;
}

static void
gtk_go_board_marshaller_1 (GtkObject     *object,
			   GtkSignalFunc func,
			   gpointer      func_data,
			   GtkArg       *args)
{
  GtkGoBoardSignal1 rfunc;

  rfunc = (GtkGoBoardSignal1) func;
  (* rfunc) (object, GTK_VALUE_INT (args[0]), GTK_VALUE_INT (args[1]), func_data);
}

static void
gtk_go_board_marshaller_2 (GtkObject     *object,
			   GtkSignalFunc func,
			   gpointer      func_data,
			   GtkArg       *args)
{
  GtkGoBoardSignal2 rfunc;

  rfunc = (GtkGoBoardSignal2) func;
  (* rfunc) (object, GTK_VALUE_INT (args[0]), GTK_VALUE_INT (args[1]),
	GTK_VALUE_INT (args[2]), func_data);
}


static void
gtk_go_board_init (GtkGoBoard *go_board)
{
  GtkWidget *widget;

  widget = GTK_WIDGET (go_board);

  go_board->high_contrast = FALSE;
  go_board->show_coordinates = FALSE;
  go_board->button_down = FALSE;

  go_board->coord_style = GTK_GO_BOARD_NORMAL_COORD;
  
  go_board->solid_stone_mask = NULL;
  go_board->trans_stone_mask = NULL;
  go_board->black_stone_pixmap = NULL;
  go_board->white_stone_pixmap = NULL;

  go_board->grey_colors = NULL;

  go_board->pos_x = -1;
  go_board->pos_y = -1;

  gtk_widget_set_events (widget,
			 gtk_widget_get_events (widget)
			 | GDK_BUTTON_PRESS_MASK
			 | GDK_BUTTON_RELEASE_MASK
			 | GDK_POINTER_MOTION_MASK
			 | GDK_LEAVE_NOTIFY_MASK);
}


GtkWidget *
gtk_go_board_new (gint x_dim,
		  gint y_dim)
{
  GtkGoBoard *go_board;

  g_return_val_if_fail (x_dim >= 2, NULL);
  g_return_val_if_fail (y_dim >= 2, NULL);

  go_board = gtk_type_new (gtk_go_board_get_type ());

  go_board->x_dim = x_dim;
  go_board->y_dim = y_dim;

  go_board->cells = NULL;
  gtk_go_board_init_cells (go_board);

  return GTK_WIDGET (go_board);
}

void
gtk_go_board_set_high_contrast (GtkGoBoard *go_board,
				gint 	    high_contrast)
{
  g_return_if_fail (go_board != NULL);
  g_return_if_fail (GTK_IS_GO_BOARD (go_board));

  if (go_board->high_contrast != high_contrast)
    {
      go_board->high_contrast = high_contrast;
      if (GTK_WIDGET_REALIZED (GTK_WIDGET (go_board)))
	{
	  gtk_go_board_update_pixmap (go_board);
	  gtk_widget_queue_draw (GTK_WIDGET (go_board));
	}
    }
}

void
gtk_go_board_set_show_coordinates (GtkGoBoard *go_board,
				   gint        show_coordinates)
{
  int w, h;
  
  g_return_if_fail (go_board != NULL);
  g_return_if_fail (GTK_IS_GO_BOARD (go_board));

  if (go_board->show_coordinates != show_coordinates)
    {
      gdk_window_get_geometry (GTK_WIDGET (go_board)->window,
			       NULL, NULL, &w, &h, NULL);

      go_board->show_coordinates = show_coordinates;
      gtk_go_board_update_layout (go_board, w, h);

      if (GTK_WIDGET_REALIZED (GTK_WIDGET (go_board)))
	{
	  gtk_go_board_update_pixmap (go_board);
	  gtk_widget_queue_draw (GTK_WIDGET (go_board));
	}
    }
}

void
gtk_go_board_set_coord_style (GtkGoBoard *go_board,
			      GtkGoBoardCoordStyle style)
{
  go_board->coord_style = style;
}

void
gtk_go_board_set_stone (GtkGoBoard *go_board,
			gint 	    x,
			gint        y,
			guint 	    stone)
{
  GdkRectangle area;

  g_return_if_fail (go_board != NULL);
  g_return_if_fail (GTK_IS_GO_BOARD (go_board));
  g_return_if_fail ((x >= 1) && (x <= go_board->x_dim));
  g_return_if_fail ((y >= 1) && (y <= go_board->y_dim));

  x--;
  y = go_board->y_dim - y;

  go_board->cells[(go_board->x_dim * y) + x].stone = stone;

  area.x = go_board->board_rect.x + (x * go_board->cell_size);
  area.y = go_board->board_rect.y + (y * go_board->cell_size);
  area.width = go_board->cell_size;
  area.height = go_board->cell_size;

  gtk_go_board_paint (GTK_WIDGET(go_board), &area);
}

guint
gtk_go_board_get_stone (GtkGoBoard *go_board,
			gint 	    x,
			gint        y)
{
  g_return_val_if_fail (go_board != NULL, 0);
  g_return_val_if_fail (GTK_IS_GO_BOARD (go_board), 0);
  g_return_val_if_fail ((x >= 1) && (x <= go_board->x_dim), 0);
  g_return_val_if_fail ((y >= 1) && (y <= go_board->y_dim), 0);

  x --;
  y = go_board->y_dim - y;

  return go_board->cells[(go_board->x_dim * y) + x].stone;
}

void
gtk_go_board_set_markup (GtkGoBoard *go_board,
			 gint        x,
			 gint        y,
			 guint       markup)
{
  GdkRectangle area;

  g_return_if_fail (go_board != NULL);
  g_return_if_fail (GTK_IS_GO_BOARD (go_board));
  g_return_if_fail ((x >= 1) && (x <= go_board->x_dim));
  g_return_if_fail ((y >= 1) && (y <= go_board->y_dim));

  x --;
  y = go_board->y_dim - y;

  if (go_board->cells[(go_board->x_dim * y) + x].markup != markup)
    {
      go_board->cells[(go_board->x_dim * y) + x].markup = markup;

      area.x = go_board->board_rect.x + (x * go_board->cell_size);
      area.y = go_board->board_rect.y + (y * go_board->cell_size);
      area.width = go_board->cell_size;
      area.height = go_board->cell_size;

      gtk_go_board_paint (GTK_WIDGET(go_board), &area);
    }
}

guint
gtk_go_board_get_markup (GtkGoBoard *go_board,
			 gint        x,
			 gint        y)
{
  g_return_val_if_fail (go_board != NULL, 0);
  g_return_val_if_fail (GTK_IS_GO_BOARD (go_board), 0);
  g_return_val_if_fail ((x >= 1) && (x <= go_board->x_dim), 0);
  g_return_val_if_fail ((y >= 1) && (y <= go_board->y_dim), 0);

  x --;
  y = go_board->y_dim - y;

  return go_board->cells[(go_board->x_dim * y) + x].markup;
}


void
gtk_go_board_add_markup (GtkGoBoard *go_board,
			 gint        x,
			 gint        y,
			 guint       markup)
{
  GdkRectangle area;

  g_return_if_fail (go_board != NULL);
  g_return_if_fail (GTK_IS_GO_BOARD (go_board));
  g_return_if_fail ((x >= 1) && (x <= go_board->x_dim));
  g_return_if_fail ((y >= 1) && (y <= go_board->y_dim));

  x --;
  y = go_board->y_dim - y;

  if ((go_board->cells[(go_board->x_dim * y) + x].markup | markup)
      != go_board->cells[(go_board->x_dim * y) + x].markup)
    {
      go_board->cells[(go_board->x_dim * y) + x].markup |= markup;

      area.x = go_board->board_rect.x + (x * go_board->cell_size);
      area.y = go_board->board_rect.y + (y * go_board->cell_size);
      area.width = go_board->cell_size;
      area.height = go_board->cell_size;

      gtk_go_board_paint (GTK_WIDGET(go_board), &area);
    }
}

void
gtk_go_board_delete_markup (GtkGoBoard *go_board,
			    gint        x,
			    gint        y,
			    guint       markup)
{
  GdkRectangle area;
  
  g_return_if_fail (go_board != NULL);
  g_return_if_fail (GTK_IS_GO_BOARD (go_board));
  g_return_if_fail ((x >= 1) && (x <= go_board->x_dim));
  g_return_if_fail ((y >= 1) && (y <= go_board->y_dim));

  x --;
  y = go_board->y_dim - y;

  if ((go_board->cells[(go_board->x_dim * y) + x].markup & ~markup)
      != go_board->cells[(go_board->x_dim * y) + x].markup)
    {
      go_board->cells[(go_board->x_dim * y) + x].markup &= ~markup;

      area.x = go_board->board_rect.x + (x * go_board->cell_size);
      area.y = go_board->board_rect.y + (y * go_board->cell_size);
      area.width = go_board->cell_size;
      area.height = go_board->cell_size;

      gtk_go_board_paint (GTK_WIDGET(go_board), &area);
    }
}

void
gtk_go_board_set_mark (GtkGoBoard *go_board,
			    gint        x,
			    gint        y,
			    guint       markup)
{
  GdkRectangle area;
  
  g_return_if_fail (go_board != NULL);
  g_return_if_fail (GTK_IS_GO_BOARD (go_board));
  g_return_if_fail ((x >= 1) && (x <= go_board->x_dim));
  g_return_if_fail ((y >= 1) && (y <= go_board->y_dim));

  x --;
  y = go_board->y_dim - y;

  if (markup != go_board->cells[(go_board->x_dim * y) + x].markup)
    {
      go_board->cells[(go_board->x_dim * y) + x].markup = markup;

      area.x = go_board->board_rect.x + (x * go_board->cell_size);
      area.y = go_board->board_rect.y + (y * go_board->cell_size);
      area.width = go_board->cell_size;
      area.height = go_board->cell_size;

      gtk_go_board_paint (GTK_WIDGET(go_board), &area);
    }
}

void
gtk_go_board_set_label (GtkGoBoard  *go_board,
			gint         x,
			gint         y,
			const gchar *label)
{
  GdkRectangle area;

  g_return_if_fail (go_board != NULL);
  g_return_if_fail (GTK_IS_GO_BOARD (go_board));
  g_return_if_fail ((x >= 1) && (x <= go_board->x_dim));
  g_return_if_fail ((y >= 1) && (y <= go_board->y_dim));

  x --;
  y = go_board->y_dim - y;

  if (go_board->cells[(go_board->x_dim * y) + x].label)
    g_free (go_board->cells[(go_board->x_dim * y) + x].label);

  if (!label || !*label)
    go_board->cells[(go_board->x_dim * y) + x].label = NULL;
  else if (strlen (label) > 0)
    go_board->cells[(go_board->x_dim * y) + x].label = g_strdup (label);
  else 
    go_board->cells[(go_board->x_dim * y) + x].label = NULL;
  go_board->cells[(go_board->x_dim * y) + x].labelcolor = 0;
  
  area.x = go_board->board_rect.x + (x * go_board->cell_size);
  area.y = go_board->board_rect.y + (y * go_board->cell_size);
  area.width = go_board->cell_size;
  area.height = go_board->cell_size;

  gtk_go_board_paint (GTK_WIDGET(go_board), &area);
}

void gtk_go_board_set_plabel (GtkGoBoard  *go_board,
			gint x, gint y, gint point)
{
  GdkRectangle area;
  gchar buf[10];
  gint color = 1;

  g_return_if_fail (go_board != NULL);
  g_return_if_fail (GTK_IS_GO_BOARD (go_board));
  g_return_if_fail ((x >= 1) && (x <= go_board->x_dim));
  g_return_if_fail ((y >= 1) && (y <= go_board->y_dim));

  x --;
  y = go_board->y_dim - y;

  if (go_board->cells[(go_board->x_dim * y) + x].label)
    g_free (go_board->cells[(go_board->x_dim * y) + x].label);

  if (point < 0) {
    color = 2;
    point = abs(point);
  }
  sprintf(buf, "%d", point);
  
  if (strlen (buf) > 0)
    go_board->cells[(go_board->x_dim * y) + x].label = g_strdup (buf);
  else 
    go_board->cells[(go_board->x_dim * y) + x].label = NULL;

  go_board->cells[(go_board->x_dim * y) + x].labelcolor = color;
  
  area.x = go_board->board_rect.x + (x * go_board->cell_size);
  area.y = go_board->board_rect.y + (y * go_board->cell_size);
  area.width = go_board->cell_size;
  area.height = go_board->cell_size;

  gtk_go_board_paint (GTK_WIDGET(go_board), &area);
}

gchar *
gtk_go_board_get_label (GtkGoBoard *go_board,
			gint        x,
			gint        y)
{
  g_return_val_if_fail (go_board != NULL, 0);
  g_return_val_if_fail (GTK_IS_GO_BOARD (go_board), 0);
  g_return_val_if_fail ((x >= 1) && (x <= go_board->x_dim), 0);
  g_return_val_if_fail ((y >= 1) && (y <= go_board->y_dim), 0);

  x --;
  y = go_board->y_dim - y;

  return go_board->cells[(go_board->x_dim * y) + x].label;
}

gint
gtk_go_board_add_line (GtkGoBoard *go_board,
		       gint    	   sx,
		       gint    	   sy,
		       gint    	   ex,
		       gint    	   ey)
{
  return TRUE;
}

void
gtk_go_board_delete_line (GtkGoBoard *go_board,
			  gint 	      tag)
{
}

gint
gtk_go_board_add_arrow (GtkGoBoard *go_board,
			gint 	    sx,
			gint        sy,
			gint        ex,
			gint        ey)
{
  return TRUE;
}

void
gtk_go_board_delete_arrow (GtkGoBoard *go_board,
			   gint        tag)
{
}


void
gtk_go_board_pressed (GtkGoBoard *go_board, gint x, gint y, gint bt)
{
  gtk_signal_emit (GTK_OBJECT(go_board), go_board_signals[PRESSED], x, y, bt);
}


void
gtk_go_board_released (GtkGoBoard *go_board, gint x, gint y, gint bt)
{
  gtk_signal_emit (GTK_OBJECT(go_board), go_board_signals[RELEASED], x, y, bt);
}

void
gtk_go_board_clicked (GtkGoBoard *go_board,
		      gint 	  x,
		      gint 	  y)
{
  gtk_signal_emit (GTK_OBJECT(go_board), go_board_signals[CLICKED], x, y);
}


void
gtk_go_board_enter (GtkGoBoard *go_board,
		    gint 	x,
		    gint 	y)
{
  gtk_signal_emit (GTK_OBJECT(go_board), go_board_signals[ENTER], x, y);
}

void
gtk_go_board_leave (GtkGoBoard *go_board,
		    gint        x,
		    gint	y)
{
  gtk_signal_emit (GTK_OBJECT(go_board), go_board_signals[LEAVE], x, y);
}



static gint
gtk_go_board_button_press (GtkWidget 	  *widget,
			   GdkEventButton *event)
{
  GtkGoBoard *go_board;
  int x, y;

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_GO_BOARD (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  if (event->type == GDK_BUTTON_PRESS)
    {
/*      if (event->button == 1) */
	{
	  go_board = GTK_GO_BOARD (widget);

	  if (gtk_go_board_get_board_coordinate (go_board,
						 event->x, event->y,
						 &x, &y))
	    {
	      go_board->press_x = x;
	      go_board->press_y = y;

	      gtk_go_board_pressed (go_board, x, y, event->button);
	    }
	}
    }

  return TRUE;
}

static gint
gtk_go_board_button_release (GtkWidget 	  *widget,
			     GdkEventButton *event)
{
  GtkGoBoard *go_board;
  int x, y;

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_GO_BOARD (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  if (event->type == GDK_BUTTON_RELEASE)
    {
/*      if (event->button == 1) */
	{
	  go_board = GTK_GO_BOARD (widget);

	  if (gtk_go_board_get_board_coordinate (go_board,
						 event->x, event->y,
						 &x, &y))
	    {
	      gtk_go_board_released (go_board, x, y, event->button);

	      if ((event->button == 1) && 
	      		(x == go_board->press_x) && (y == go_board->press_y))
		gtk_go_board_clicked (go_board, x, y);
	    }

	  go_board->press_x = go_board->press_y = -1;
	}
    }
  return TRUE;
}


static gboolean
gtk_go_board_get_board_coordinate (GtkGoBoard *go_board,
				   int x, int y,
				   int *rx, int *ry)
{
  if ((x < go_board->board_rect.x)
      || (x >= (go_board->board_rect.x + go_board->board_rect.width))
      || (y < go_board->board_rect.y)
      || (y >= (go_board->board_rect.y + go_board->board_rect.height)))
    return FALSE;

  if (rx)
    *rx = ((x - go_board->board_rect.x) / go_board->cell_size) + 1;
  if (ry)
    *ry = go_board->y_dim -
      ((y - go_board->board_rect.y) / go_board->cell_size);

  return TRUE;
}


static gint
gtk_go_board_leave_notify (GtkWidget        *widget,
			   GdkEventCrossing *event)
{
  GtkGoBoard *go_board;

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_GO_BOARD (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  go_board = GTK_GO_BOARD (widget);

  if (go_board->pos_x > 0 && go_board->pos_y > 0)
    {
      gtk_go_board_leave (go_board, go_board->pos_x, go_board->pos_y);
      go_board->pos_x = -1;
      go_board->pos_y = -1;
    }

  return TRUE;
}

static gint
gtk_go_board_motion_notify (GtkWidget      *widget,
			    GdkEventMotion *event)
{
  int x, y;
  gboolean entered;
  GtkGoBoard *go_board;

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_GO_BOARD (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  go_board = GTK_GO_BOARD (widget);

  entered = gtk_go_board_get_board_coordinate (go_board,
					       event->x, event->y,
					       &x, &y);

  if (entered && go_board->pos_x == x && go_board->pos_y == y)
    return TRUE;

  if (go_board->pos_x > 0 && go_board->pos_y > 0)
    {
      gtk_go_board_leave (go_board, go_board->pos_x, go_board->pos_y);
      go_board->pos_x = -1;
      go_board->pos_y = -1;
    }

  if (entered)
    {
      gtk_go_board_enter (go_board, x, y);
      go_board->pos_x = x;
      go_board->pos_y = y;
    }

  return TRUE;
}

static gint
gtk_go_board_expose (GtkWidget      *widget,
		     GdkEventExpose *event)
{
  GtkGoBoard *go_board;

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_GO_BOARD (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  if (GTK_WIDGET_DRAWABLE (widget))
    {
      go_board = GTK_GO_BOARD (widget);

      gtk_go_board_paint (widget, &event->area);
    }
  return TRUE;
}

static void
gtk_go_board_destroy (GtkObject *object)
{
  g_return_if_fail (object != NULL);
  g_return_if_fail (GTK_IS_GO_BOARD (object));
}

static void
gtk_go_board_finalize (GtkObject *object)
{
  GtkGoBoard *go_board;

  g_return_if_fail (object != NULL);
  g_return_if_fail (GTK_IS_GO_BOARD (object));

  go_board = GTK_GO_BOARD (object);

  /* free pixmaps. */
  if (go_board->solid_stone_mask)
    gdk_pixmap_unref (go_board->solid_stone_mask);
  if (go_board->trans_stone_mask)
    gdk_pixmap_unref (go_board->solid_stone_mask);
  if (go_board->black_stone_pixmap)
    gdk_pixmap_unref (go_board->black_stone_pixmap);
  if (go_board->white_stone_pixmap)
    gdk_pixmap_unref (go_board->white_stone_pixmap);

  /* free grey colors. */
  gtk_go_board_greys_free (go_board);
}

static void
gtk_go_board_size_request (GtkWidget      *widget,
			   GtkRequisition *requisition)
{
  GtkGoBoard *go_board;

  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_GO_BOARD (widget));
  g_return_if_fail (requisition != NULL);

  go_board = GTK_GO_BOARD (widget);

  requisition->width = DEFAULT_CELL_SIZE * go_board->x_dim;
  requisition->height = DEFAULT_CELL_SIZE * go_board->y_dim;

  if (go_board->show_coordinates)
    {
      requisition->width += 2 * gtk_go_board_max_coord_width (go_board);
      requisition->height += 2 * gtk_go_board_max_coord_height (go_board);
    }
}

static void
gtk_go_board_size_allocate (GtkWidget    	*widget,
			    GtkAllocation	*allocation)
{
  GtkGoBoard *go_board;

  go_board = GTK_GO_BOARD (widget);

  (* parent_size_allocate) (widget, allocation);
  gtk_go_board_update_layout (go_board,
			      allocation->width, allocation->height);
  
  if (GTK_WIDGET_REALIZED (widget))
    gtk_go_board_update_pixmap (go_board);
}


static void
gtk_go_board_realize (GtkWidget	*widget)
{
  GtkGoBoard *go_board;
  int w, h;

  go_board = GTK_GO_BOARD (widget);

  (* parent_realize) (widget);


  gdk_window_get_geometry (widget->window,
			   NULL, NULL, &w, &h, NULL);

  gtk_go_board_update_layout (go_board, w, h);
  gtk_go_board_update_pixmap (go_board);
}


/*
   * Each cell is a square.

   +-----------------------+
   |	dx		   |
   |dy +--------------+    |
   |   |   x-coords   |    |
   |   |  +--------+  |    |
   |   |y |  | 	|  |y |    |
   |   || |--O--+--|| |    |
   |   |c |  | 	|  |c |    |
   |   |o |--+--+--|o |    |
   |   |r |  | 	|  |r |    |
   |   |d |--+--@--|d |    |
   |   |s |  | 	|  |s |    |
   |   |  +--------+  |    |
   |   |   x-coords   |    |
   |   +--------------+    |
   +-----------------------+
 */

static void
gtk_go_board_update_layout (GtkGoBoard *go_board, gint width, gint height)
{
  GtkWidget *widget;
  gint fw = 0, fh = 0;

  gint x_dim, y_dim, cell_size;

  x_dim = go_board->x_dim;
  y_dim = go_board->y_dim;
  
  if (go_board->show_coordinates)
    {
      fh = gtk_go_board_max_coord_height (go_board);
      fw = gtk_go_board_max_coord_width (go_board);

      widget = GTK_WIDGET (go_board);

      cell_size = MIN ((width - 2 * fw) / x_dim, (height - 2 * fh) / y_dim);
    }
  else
    {
      cell_size = MIN (width / x_dim, height / y_dim);
    }

  go_board->cell_size = cell_size;

  go_board->board_rect.x = (width - (cell_size * x_dim)) / 2;
  go_board->board_rect.width = cell_size * x_dim;
  go_board->board_rect.y = (height - (cell_size * y_dim)) / 2;
  go_board->board_rect.height = cell_size * y_dim;

  if (go_board->show_coordinates)
    {
      go_board->upper_coordinates_rect.x = go_board->board_rect.x;
      go_board->upper_coordinates_rect.width = go_board->board_rect.width;
      go_board->upper_coordinates_rect.y = go_board->board_rect.y - fh;
      go_board->upper_coordinates_rect.height = fh;

      go_board->lower_coordinates_rect.x = go_board->board_rect.x;
      go_board->lower_coordinates_rect.width = go_board->board_rect.width;
      go_board->lower_coordinates_rect.y
	= go_board->board_rect.y + go_board->board_rect.height;
      go_board->lower_coordinates_rect.height = fh;

      go_board->left_coordinates_rect.x = go_board->board_rect.x - fw;
      go_board->left_coordinates_rect.width = fw;
      go_board->left_coordinates_rect.y = go_board->board_rect.y;
      go_board->left_coordinates_rect.height = go_board->board_rect.height;

      go_board->right_coordinates_rect.x
	= go_board->board_rect.x + go_board->board_rect.width;
      go_board->right_coordinates_rect.width = fw;
      go_board->right_coordinates_rect.y
	= go_board->board_rect.y;
      go_board->right_coordinates_rect.height = go_board->board_rect.height;
    }
}




static void
gtk_go_board_update_pixmap (GtkGoBoard *go_board)
{
  GtkWidget *widget;
  GdkWindow *window;
  int size = go_board->cell_size;
  int max_radius = size * size;
  float bright, lambertian, z;
  int b_bright, w_bright;
  int x, y, mask_width;
  GdkGC *gc;

  guchar *solid_mask_data, *trans_mask_data;
/*  GdkColor color; */

  widget = GTK_WIDGET (go_board);
  window = widget->window;

  /* update masks */
  mask_width = (size + 7) >> 3;
  solid_mask_data = g_malloc0 (mask_width * size);
  trans_mask_data = g_malloc0 (mask_width * size);

  if (go_board->solid_stone_mask)
    gdk_pixmap_unref (go_board->solid_stone_mask);
  if (go_board->trans_stone_mask)
    gdk_pixmap_unref (go_board->trans_stone_mask);
  for (y = 0; y < size; y++)
    for (x = 0; x < size; x++)
      {
	if (((size - 1) - (x+x)) * ((size - 1) - (x+x)) +
	    ((size - 1) - (y+y)) * ((size - 1) - (y+y)) <= max_radius)
	  {
	    solid_mask_data[(x >> 3) + (y * mask_width)] |= (1 << (x & 7));
	    if ((x+y) % 2)
	      trans_mask_data[(x >> 3) + (y * mask_width)] |= (1 << (x & 7));
	  }
      }
  go_board->solid_stone_mask
    = gdk_bitmap_create_from_data (window, solid_mask_data, size, size);
  go_board->trans_stone_mask
    = gdk_bitmap_create_from_data (window, trans_mask_data, size, size);

  g_free (solid_mask_data);
  g_free (trans_mask_data);

  /* update stone images */
  if (go_board->black_stone_pixmap)
    gdk_pixmap_unref (go_board->black_stone_pixmap);
  if (go_board->white_stone_pixmap)
    gdk_pixmap_unref (go_board->white_stone_pixmap);

  go_board->black_stone_pixmap = gdk_pixmap_new (window, size, size, -1);
  go_board->white_stone_pixmap = gdk_pixmap_new (window, size, size, -1);

  gc = widget->style->fg_gc[GTK_STATE_NORMAL];

  if (! go_board->high_contrast
      && gtk_go_board_greys_alloc (go_board))
    {
      for (y = 0; y < size; y++)
	for (x = 0; x < size; x++)
	  {
	    bright = gtk_go_board_calc_reflection ((size - 1) - (x + x),
						   (size - 1) - (y + y),
						   size + size,
						   &lambertian, &z);
	    bright *= bright;
	    bright *= bright;
	    b_bright = bright * 165.0 + lambertian * 10.0 - 5.0;
	    if (b_bright > 255)
	      b_bright = 255;
	    if (b_bright < 0)
	      b_bright = 0;

	    gdk_gc_set_foreground (gc, &(go_board->grey_colors[b_bright]));
	    gdk_draw_point (go_board->black_stone_pixmap, gc, x, y);

	    bright *= bright;
	    bright *= bright;
	    bright *= bright;
	    w_bright = bright * bright * 300.0 + lambertian * 120.0 + 110.0;
	    if (w_bright > 255)
	      w_bright = 255;
	    if (w_bright < 0)
	      w_bright = 0;

	    gdk_gc_set_foreground (gc, &(go_board->grey_colors[w_bright]));
	    gdk_draw_point (go_board->white_stone_pixmap, gc, x, y);
	  }
    }
  else
    {
      go_board->high_contrast = TRUE;
      gtk_go_board_greys_free (go_board);
      
      gdk_gc_set_foreground (gc, &(widget->style->white));
      gdk_draw_rectangle (go_board->white_stone_pixmap, gc, TRUE,
			  0, 0, size, size);
      gdk_gc_set_foreground (gc, &(widget->style->black));
      max_radius = (size - 2) * (size - 2);
      for (y = 0; y < size; y++)
	for (x = 0; x < size; x++)
	  {
	    if (((size - 1) - (x+x)) * ((size - 1) - (x+x)) +
		((size - 1) - (y+y)) * ((size - 1) - (y+y)) >= max_radius)
	      gdk_draw_point (go_board->white_stone_pixmap, gc, x, y);
	  }
      gdk_draw_rectangle (go_board->black_stone_pixmap, gc, TRUE,
			  0, 0, size, size);
    }
  gdk_gc_set_foreground (gc, &(widget->style->fg[GTK_STATE_NORMAL]));
}

GdkColor colorblue = {0, 1000, 30000, 60000};        
GdkColor colorred = {0, 50000, 10000, 55000};

static void
gtk_go_board_paint (GtkWidget    *widget,
		    GdkRectangle *area)
{
  GtkGoBoard *go_board;
  int x_dim, y_dim, cell_size;
  guint stone, markup, lineinfo, labelcolor;
  gchar *label;


  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_GO_BOARD (widget));

  if (GTK_WIDGET_DRAWABLE (widget))
    {
      int delta_x, delta_y;
      GdkWindow *window;
      GdkRectangle destrect;
      GdkGC *gc;

      go_board = GTK_GO_BOARD (widget);
      window = widget->window;
      gc = widget->style->fg_gc[GTK_STATE_NORMAL];

      gdk_window_clear_area (widget->window,
			     area->x, area->y,
			     area->width, area->height);


      x_dim = go_board->x_dim;
      y_dim = go_board->y_dim;
      
      cell_size = go_board->cell_size;
      delta_x = go_board->board_rect.x;
      delta_y = go_board->board_rect.y;


      /* draw coordinates. */
      if (go_board->show_coordinates)
	{
	  GdkFont *font = widget->style->font;
	  GdkRectangle *rect;
	  int i, w, h;
	  char *string;

	  /* FIXME: fix this to redraw the required area! */
	  rect = &(go_board->upper_coordinates_rect);
	  if (gdk_rectangle_intersect (area, rect, &destrect))
	    {
	      for (i = 0; i < x_dim; i++)
		{
		  string = gtk_go_board_row_coord_string (go_board, i);
		  w = gdk_string_width (font, string);
		  gdk_draw_string (window, font, gc,
				   rect->x + i * cell_size + cell_size / 2
				   - w / 2,
				   rect->y + font->ascent,
				   string);
		}
	    }
	  
	  rect = &(go_board->lower_coordinates_rect);
	  if (gdk_rectangle_intersect (area, rect, &destrect))
	    {
	      for (i = 0; i < x_dim; i++)
		{
		  string = gtk_go_board_row_coord_string (go_board, i);
		  w = gdk_string_width (font, string);
		  gdk_draw_string (window, font, gc,
				   rect->x + i * cell_size + cell_size / 2
				   - w / 2,
				   rect->y + font->ascent,
				   string);
		}
	    }

	  rect = &(go_board->left_coordinates_rect);
	  if (gdk_rectangle_intersect (area, rect, &destrect))
	    {
	      for (i = 0; i < y_dim; i++)
		{
		  string = gtk_go_board_col_coord_string (go_board, i);
		  w = gdk_string_width (font, string);
		  h = font->ascent + font->descent;
		  gdk_draw_string (window, font, gc,
				   rect->x + (rect->width - w) / 2,
				   rect->y + i * cell_size + cell_size / 2
				   + font->ascent - h / 2,
				   string);
		}
	    }

	  rect = &(go_board->right_coordinates_rect);
	  if (gdk_rectangle_intersect (area, rect, &destrect))
	    {
	      for (i = 0; i < y_dim; i++)
		{
		  string = gtk_go_board_col_coord_string (go_board, i);
		  w = gdk_string_width (font, string);
		  h = font->ascent + font->descent;
		  gdk_draw_string (window, font, gc,
				   rect->x + (rect->width - w) / 2,
				   rect->y + i * cell_size + cell_size / 2
				   + font->ascent - h / 2,
				   string);
		}
	    }
	}

      /* draw board */
      if (gdk_rectangle_intersect (area, &(go_board->board_rect), &destrect))
	{
	  int sx, sy, ex, ey, x, y, i;
	  int px1 = 0, px2 = 0, py1 = 0, py2 = 0;

	  sx = MAX((area->x - delta_x) / cell_size, 0);
	  sy = MAX((area->y - delta_y) / cell_size, 0);
	  ex = MIN((area->x + area->width - 1 - delta_x) / cell_size,
		   x_dim - 1);
	  ey = MIN((area->y + area->height - 1- delta_y) / cell_size,
		   y_dim - 1);


	  for (y = sy; y <= ey; y++)
	    {
	      for (x = sx; x <= ex; x++)
		{
		  stone = go_board->cells[y * x_dim + x].stone;
		  markup = go_board->cells[y * x_dim + x].markup;
		  label = go_board->cells[y * x_dim + x].label;
		  labelcolor = go_board->cells[y * x_dim + x].labelcolor;
		  lineinfo = go_board->cells[y * x_dim + x].lineinfo;

		  if ((! IS_NO_LINE_MARKUP(markup)) && (! label))
		    {
		      /* draw line. */
		      gdk_gc_set_foreground (gc, &widget->style->black);
		      gdk_gc_set_line_attributes (gc,
						  1, GDK_LINE_SOLID,
						  GDK_CAP_ROUND, GDK_JOIN_ROUND);
		      switch (lineinfo & HORIZONTAL_LINE)
			{
			case HORIZONTAL_FULL_LINE:
			  px1 = delta_x + (x * cell_size);
			  px2 = delta_x + ((x + 1) * cell_size) - 1;
			  break;
			case HORIZONTAL_RIGHT_LINE:
			  px1 = delta_x + (cell_size / 2);
			  px2 = delta_x + cell_size - 1;
			  break;
			case HORIZONTAL_LEFT_LINE:
			  px1 = delta_x + (x * cell_size);
			  px2 = delta_x + (x * cell_size) + (cell_size / 2);
			  break;
			}

		      py1 = delta_y + (y * cell_size) + (cell_size / 2);
		      gdk_draw_line ((GdkDrawable *) widget->window, gc,
				     px1, py1, px2, py1);

		      switch (lineinfo & VERTICAL_LINE)
			{
			case VERTICAL_FULL_LINE:
			  py1 = delta_y + (y * cell_size);
			  py2 = delta_y + ((y + 1) * cell_size) - 1;
			  break;
			case VERTICAL_LOWER_LINE:
			  py1 = delta_y + (cell_size / 2);
			  py2 = delta_y + cell_size - 1;
			  break;
			case VERTICAL_UPPER_LINE:
			  py1 = delta_y + (y * cell_size);
			  py2 = delta_y + (y * cell_size) + (cell_size / 2);
			  break;
			}

		      px1 = delta_x + (x * cell_size) + (cell_size / 2);

		      gdk_draw_line ((GdkDrawable *) widget->window, gc, 
				     px1, py1, px1, py2);

		      /* draw handicap point. */
		      if (go_board->cells[y*x_dim + x].lineinfo & HANDICAP_POINT)
			{
			  px1 = delta_x + (x * cell_size) + (cell_size / 2) - (cell_size / 8);
			  py1 = delta_y + (y * cell_size) + (cell_size / 2) - (cell_size / 8);
			  px2 = (cell_size / 4);

			  gdk_draw_arc ((GdkDrawable *) widget->window, gc,
					TRUE,
					px1, py1,
					px2, px2,
					0, 23040);
			}
		    }

		  /* draw stone. */
		  if (! GTK_GO_BOARD_IS_EMPTY (stone))
		    {
		      if (GTK_GO_BOARD_IS_SOLID (stone))
			gdk_gc_set_clip_mask (gc, go_board->solid_stone_mask);
		      else
			gdk_gc_set_clip_mask (gc, go_board->trans_stone_mask);

		      gdk_gc_set_clip_origin (gc,
					      delta_x + (x * cell_size),
					      delta_y + (y * cell_size));

		      if (GTK_GO_BOARD_IS_WHITE (stone))
			gdk_window_copy_area (widget->window,
					      gc,
					      delta_x + (x * cell_size),
					      delta_y + (y * cell_size),
					      go_board->white_stone_pixmap,
					      0, 0, cell_size, cell_size);
		      else
			gdk_window_copy_area (widget->window,
					      gc,
					      delta_x + (x * cell_size),
					      delta_y + (y * cell_size),
					      go_board->black_stone_pixmap,
					      0, 0, cell_size, cell_size);
		      gdk_gc_set_clip_mask (gc, NULL);
		    }

		  /* draw markups */
		  if (markup)
		    {
		      if (IS_NO_LINE_MARKUP (markup))
			  {
			    int line_width;
			    if (GTK_GO_BOARD_IS_EMPTY (stone))
			      gdk_gc_set_foreground (gc, &widget->style->black);
			    else if (GTK_GO_BOARD_IS_BLACK (stone))
			      gdk_gc_set_foreground (gc, &widget->style->white);
			    else
			      gdk_gc_set_foreground (gc, &widget->style->black);
			  
			    line_width = cell_size / 10;
			    if (line_width < 1)
			      line_width = 1;
			    gdk_gc_set_line_attributes (gc,
							line_width,
							GDK_LINE_SOLID,
							GDK_CAP_ROUND,
							GDK_JOIN_ROUND);

			    gdk_gc_set_clip_mask (gc, go_board->solid_stone_mask);
			    gdk_gc_set_clip_origin (gc,
						    delta_x + (x * cell_size),
						    delta_y + (y * cell_size));

			    if (GTK_GO_BOARD_IS_MA (markup))
			      {
				px1 =  delta_x + (x * cell_size) + (cell_size / 4);
				py1 =  delta_y + (y * cell_size) + (cell_size / 4);
				px2 =  delta_x + (x * cell_size) + (cell_size / 4) * 3;
				py2 =  delta_y + (y * cell_size) + (cell_size / 4) * 3;
				gdk_draw_line (widget->window, gc,
					       px1, py1, px2, py2);
				px1 ^= px2;
				px2 ^= px1;
				px1 ^= px2;
				gdk_draw_line (widget->window, gc,
					       px1, py1, px2, py2);
			      }
			    else if (GTK_GO_BOARD_IS_TR (markup))
			      {
				px1 =  delta_x + (x * cell_size) + (cell_size / 2);
				py1 =  delta_y + (y * cell_size) + (cell_size / 9);
				px2 =  delta_x + (x * cell_size) + (cell_size / 9);
				py2 =  delta_y + (y * cell_size)
				  + (cell_size / 4) * 3;

				gdk_draw_line (widget->window, gc,
					       px1, py1, px2, py2);
				px2 = delta_x + (x * cell_size) + cell_size
				  - (cell_size / 9);
				gdk_draw_line (widget->window, gc,
					       px1, py1, px2, py2);
				px1 = delta_x + (x * cell_size) + (cell_size / 9);
				py1 = py2;
				gdk_draw_line (widget->window, gc,
					       px1, py1, px2, py2);
			  
			      }
			    else if (GTK_GO_BOARD_IS_CR (markup))
			      {
				px1 =  delta_x + (x * cell_size) + (cell_size / 4);
				py1 =  delta_y + (y * cell_size) + (cell_size / 4);
				gdk_draw_arc (widget->window,	gc, FALSE,
					      px1, py1,
					      cell_size / 2, cell_size / 2,
					      0, 23040);
			      }
		      
			    if (GTK_GO_BOARD_IS_SQ (markup))
			      {
				px1 =  delta_x + (x * cell_size) + (cell_size / 4);
				py1 =  delta_y + (y * cell_size) + (cell_size / 4);
				px2 =  delta_x + (x * cell_size) + (cell_size / 4) * 3;
				py2 =  delta_y + (y * cell_size) + (cell_size / 4) * 3;
				gdk_draw_rectangle (widget->window, gc, FALSE,
						    px1, py1, 
						    cell_size / 2, cell_size / 2);
			      }
			    else if (GTK_GO_BOARD_IS_SL (markup))
			      {
				px1 =  delta_x + (x * cell_size) + (cell_size / 3);
				py1 =  delta_y + (y * cell_size) + (cell_size / 3);
				gdk_draw_rectangle (widget->window, gc, TRUE,
						    px1, py1,
						    cell_size / 3, cell_size / 3);
			      }
			    gdk_gc_set_clip_mask (gc, NULL);
			  }
		      
		      if (GTK_GO_BOARD_IS_SLASH (markup))
			{
			  if (GTK_GO_BOARD_IS_WHITE_SLASH (markup))
			      gdk_gc_set_foreground (gc, &widget->style->white);
			  else
			      gdk_gc_set_foreground (gc, &widget->style->black);

			  gdk_gc_set_line_attributes (gc,
						      1,
						      GDK_LINE_SOLID,
						      GDK_CAP_ROUND,
						      GDK_JOIN_ROUND);

			  px1 = delta_x + (x * cell_size);
			  py1 = delta_y + (y * cell_size);
			  px2 = delta_x + (x * cell_size) + cell_size - 1;
			  py2 = delta_y + (y * cell_size) + cell_size - 1;

			  for (i = 0; i < cell_size; i += 4)
			    {
			      gdk_draw_line (widget->window, gc,
					     px1 + i, py1, px2, py2 - i);
			      gdk_draw_line (widget->window, gc,
					     px1, py1 + i, px2 - i, py2);
			    }
			}
		    }
		  
		  if (label)
		    {
		      int fw, fh;
		      GdkFont *font = widget->style->font;

		      if (labelcolor) {
		        if (labelcolor == 1)
		          gdk_gc_set_foreground (gc, &colorblue);
		        else
		          gdk_gc_set_foreground (gc, &colorred);
		      } else if (GTK_GO_BOARD_IS_EMPTY (stone))
			gdk_gc_set_foreground (gc, &widget->style->black);
		      else if (GTK_GO_BOARD_IS_BLACK (stone))
			gdk_gc_set_foreground (gc, &widget->style->white);
		      else
			gdk_gc_set_foreground (gc, &widget->style->black);

		      gdk_gc_set_clip_mask (gc, go_board->solid_stone_mask);
		      gdk_gc_set_clip_origin (gc,
					      delta_x + (x * cell_size),
					      delta_y + (y * cell_size));
		      
		      fw = gdk_string_width (font, label);
		      fh = font->ascent + font->descent;
		      px1 = delta_x + (cell_size * x) + (cell_size / 2)
			- (fw / 2);
		      py1 = delta_y + (cell_size * y) + (cell_size / 2)
			- (fh / 2) + font->ascent;
		      
		      gdk_draw_string (widget->window, font, gc,
				       px1, py1, label);
		      gdk_gc_set_clip_mask (gc, NULL);
		    }
		}
	    }
	  gdk_gc_set_foreground (gc, &widget->style->fg[GTK_STATE_NORMAL]);
	  gdk_gc_set_line_attributes (gc,
				      1, GDK_LINE_SOLID,
				      GDK_CAP_ROUND, GDK_JOIN_ROUND);
	}
    }
}


/* Calculate stone reflection.

   Come from CGoBan, CalcCosAngleReflection2View.  */

static float
gtk_go_board_calc_reflection (int x, int y, int r,
			      float *lambertian, float *z)
{
  const float  lx = 0.35355339, ly = 0.35355339, lz = 0.8660254;
  float  nx, ny, nz, rz;
  float  n_dot_l;

  nz = (float) sqrt ((double) (r * r - x * x - y * y));
  *z = 1.0 - (nz / r);
  nx = (float) x;
  ny = (float) y;
  n_dot_l = (nx * lx + ny * ly + nz * lz) / r;
  rz = (2.0 * nz * n_dot_l) / r - lz;
  *lambertian = n_dot_l;
  return (rz);
}


/* Allocate grey colors for high-contrast stones. */

static gboolean
gtk_go_board_greys_alloc (GtkGoBoard *go_board)
{
  int i, failed;
  GdkWindow *window;
  GdkVisual *visual;
  GdkColormap *colormap;
  GdkColorContext *color_context;

  if (! go_board->grey_colors)
    {
      window = GTK_WIDGET (go_board)->window;
      visual = gdk_window_get_visual(window);

      colormap = gdk_window_get_colormap (window);
      color_context = gdk_color_context_new (visual, colormap);

      go_board->grey_colors = g_new (GdkColor, 256);
      for (i = 0; i < 256; i++)
	{
	  go_board->grey_colors[i].red = i * 256.0;
	  go_board->grey_colors[i].green = i * 256.0;
	  go_board->grey_colors[i].blue = i * 256.0;
	  go_board->grey_colors[i].pixel
	    = gdk_color_context_get_pixel (color_context,
					   go_board->grey_colors[i].red,
					   go_board->grey_colors[i].green,
					   go_board->grey_colors[i].blue,
					   &failed);
	  if (failed)
	    {
	      gtk_go_board_greys_free (go_board);
	      return FALSE;
	    }
	}

  colorblue.pixel = gdk_color_context_get_pixel(color_context,
  	colorblue.red, colorblue.green, colorblue.blue, &failed);
  colorred.pixel = gdk_color_context_get_pixel(color_context,
  	colorred.red, colorred.green, colorred.blue, &failed);

      gdk_color_context_free (color_context);
    }
  
	
  return TRUE;
}


/* Deallocate grey colors for high-contrast stones. */

static void
gtk_go_board_greys_free (GtkGoBoard *go_board)
{
/*  int i; */

  if (go_board->grey_colors)
    {
      g_free (go_board->grey_colors);
      go_board->grey_colors = NULL;
    }
}


/* Initialize various cell states and information about drawing lines.  */

static void
gtk_go_board_init_cells (GtkGoBoard *go_board)
{
  gint ix, iy;
  gint x_dim, y_dim;
  GtkGoBoardCell *cells;

  x_dim = go_board->x_dim;
  y_dim = go_board->y_dim;
  cells = go_board->cells;

  if (go_board->cells)
    g_free (go_board->cells);

  cells = g_new0 (GtkGoBoardCell, x_dim * y_dim);
  for (ix = 0; ix < x_dim; ix++)
    for (iy = 0; iy < y_dim; iy++)
      {
	if (ix == 0)
	  cells[ix + iy * x_dim].lineinfo |= HORIZONTAL_RIGHT_LINE;
	else if (ix == (x_dim - 1))
	  cells[ix + iy * x_dim].lineinfo |= HORIZONTAL_LEFT_LINE;
	else
	  cells[ix + iy * x_dim].lineinfo |= HORIZONTAL_FULL_LINE;

	if (iy == 0)
	  cells[ix + iy * x_dim].lineinfo |= VERTICAL_LOWER_LINE;
	else if (iy == (y_dim - 1))
	  cells[ix + iy * x_dim].lineinfo |= VERTICAL_UPPER_LINE;
	else
	  cells[ix + iy * x_dim].lineinfo |= VERTICAL_FULL_LINE;

	cells[ix + iy * x_dim].stone = GTK_GO_BOARD_EMPTY;
	cells[ix + iy * x_dim].label = NULL;
	cells[ix + iy * x_dim].arrows = NULL;
	cells[ix + iy * x_dim].lines = NULL;
      }

  if (x_dim == y_dim)
    {
      switch (x_dim)
	{
	case 9:
	  cells[2 * y_dim + 2].lineinfo |= HANDICAP_POINT;
	  cells[2 * y_dim + 6].lineinfo |= HANDICAP_POINT;
	  cells[4 * y_dim + 4].lineinfo |= HANDICAP_POINT;
	  cells[6 * y_dim + 2].lineinfo |= HANDICAP_POINT;
	  cells[6 * y_dim + 6].lineinfo |= HANDICAP_POINT;
	  break;
	case 13:
	  cells[3 * y_dim + 3].lineinfo |= HANDICAP_POINT;
	  cells[3 * y_dim + 6].lineinfo |= HANDICAP_POINT;
	  cells[3 * y_dim + 9].lineinfo |= HANDICAP_POINT;
	  cells[6 * y_dim + 3].lineinfo |= HANDICAP_POINT;
	  cells[6 * y_dim + 6].lineinfo |= HANDICAP_POINT;
	  cells[6 * y_dim + 9].lineinfo |= HANDICAP_POINT;
	  cells[9 * y_dim + 3].lineinfo |= HANDICAP_POINT;
	  cells[9 * y_dim + 6].lineinfo |= HANDICAP_POINT;
	  cells[9 * y_dim + 9].lineinfo |= HANDICAP_POINT;
	  break;
	case 19:
	  cells[3 * y_dim + 3].lineinfo |= HANDICAP_POINT;
	  cells[3 * y_dim + 9].lineinfo |= HANDICAP_POINT;
	  cells[3 * y_dim + 15].lineinfo |= HANDICAP_POINT;
	  cells[9 * y_dim + 3].lineinfo |= HANDICAP_POINT;
	  cells[9 * y_dim + 9].lineinfo |= HANDICAP_POINT;
	  cells[9 * y_dim + 15].lineinfo |= HANDICAP_POINT;
	  cells[15 * y_dim + 3].lineinfo |= HANDICAP_POINT;
	  cells[15 * y_dim + 9].lineinfo |= HANDICAP_POINT;
	  cells[15 * y_dim + 15].lineinfo |= HANDICAP_POINT;
	  break;
	}
    }

  go_board->cells = cells;
}

static gchar *
gtk_go_board_row_coord_string (GtkGoBoard *go_board, int n)
{
  static char buffer[16];

  if (go_board->coord_style == GTK_GO_BOARD_IGS_COORD)
    {
      if (n < ('I' - 'A'))
	sprintf (buffer, "%c", 'A' + n);
      else
	sprintf (buffer, "%c", 'A' + n + 1);
    }
  else
    sprintf (buffer, "%c", 'A' + n);
  return buffer;
}

static gchar *
gtk_go_board_col_coord_string (GtkGoBoard *go_board, int n)
{
  static char buffer[16];

  sprintf (buffer, "%d", go_board->x_dim - n);
  return buffer;
}


static gint
gtk_go_board_max_coord_height (GtkGoBoard *go_board)
{
  GdkFont *font;

  font = (GTK_WIDGET (go_board))->style->font;

  return font->ascent + font->descent;
}



static gint
gtk_go_board_max_coord_width (GtkGoBoard *go_board)
{
  GdkFont *font;
  char *string;
  int max, i;

  font = (GTK_WIDGET (go_board))->style->font;

  max = 0;
  for (i = 0; i < go_board->x_dim; i++)
    {
      string = gtk_go_board_row_coord_string (go_board, i);
      max = MAX (max, gdk_string_width (font, string));
    }
  for (i = 0; i < go_board->y_dim; i++)
    {
      string = gtk_go_board_col_coord_string (go_board, i);
      max = MAX (max, gdk_string_width (font, string));
    }

  return max;
}

void gtk_go_board_clear_board (GtkGoBoard *go_board)
{
  gint ix, iy;
  
  for (ix = 0; ix < go_board->x_dim; ix++)
    for (iy = 0; iy < go_board->y_dim; iy++)
      {
	go_board->cells[ix + iy * go_board->x_dim].stone = GTK_GO_BOARD_EMPTY;
	go_board->cells[ix + iy * go_board->x_dim].label = NULL;
	go_board->cells[ix + iy * go_board->x_dim].labelcolor = 0;
	go_board->cells[ix + iy * go_board->x_dim].arrows = NULL;
	go_board->cells[ix + iy * go_board->x_dim].lines = NULL;
	go_board->cells[ix + iy * go_board->x_dim].markup = 0;
      }
  gtk_go_board_paint((GtkWidget *) go_board, &go_board->board_rect);
}
