/* Baduki - Computer Baduk(Go) Program 
   Copyright(c) 1998-1999 Lim Jaebum <artist@soback.kornet21.net> */

/* $Id: gtkgoboard.h,v 1.5 1999/01/15 18:41:54 artist Exp $ */

/* gtkgoboard.h -- GTK+ go board widget.
   Copyright (C) 1998 Copyright (C) 1998 Changwoo Ryu

   This program is free software; you can redistribute it and'or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
   USA */

/* Written by Changwoo Ryu <cwryu@adam.kaist.ac.kr>. */

/* 
   This widget is made for displaying/playing/editing go game, which can be
   described with SGF (Smart Game Format) FF[4].

   But it has no SGF load/save feature or tree game history saving.
   Input/Output is the only purpose of this widget.

   For more information about SGF FF[4], visit
   <http://www.sbox.tu-graz.ac.at/home/hollosi/sgf/>.
*/

#ifndef __GTK_GO_BOARD_H__
#define __GTK_GO_BOARD_H__

#include <gtk/gtkdrawingarea.h>

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


#define GTK_GO_BOARD(obj)          GTK_CHECK_CAST (obj, gtk_go_board_get_type (), GtkGoBoard)
#define GTK_GO_BOARD_CLASS(klass)  GTK_CHECK_CLASS_CAST (klass, gtk_go_board_get_type (), GtkGoBoardClass)
#define GTK_IS_GO_BOARD(obj)       GTK_CHECK_TYPE (obj, gtk_go_board_get_type ())

/* stone states -- one of these

   (1) No Stone `0'
   (2) Yes Stone xx1
     1) Trans (preparing) Stone x01
       - Black `001'
       - White `101'
     2) Solid (real) Stone  x11
       - Black `011'
       - White `111'
*/     

#define GTK_GO_BOARD_NONEMPTY_MASK 	(1 << 0)
#define GTK_GO_BOARD_SOLID_MASK 	(1 << 1)
#define GTK_GO_BOARD_WHITE_MASK 	(1 << 2)
  
#define GTK_GO_BOARD_EMPTY 0
#define GTK_GO_BOARD_TRANS_BLACK (GTK_GO_BOARD_NONEMPTY_MASK)
#define GTK_GO_BOARD_TRANS_WHITE (GTK_GO_BOARD_NONEMPTY_MASK | GTK_GO_BOARD_WHITE_MASK)
#define GTK_GO_BOARD_SOLID_BLACK (GTK_GO_BOARD_NONEMPTY_MASK | GTK_GO_BOARD_SOLID_MASK)
#define GTK_GO_BOARD_SOLID_WHITE (GTK_GO_BOARD_NONEMPTY_MASK | GTK_GO_BOARD_SOLID_MASK | GTK_GO_BOARD_WHITE_MASK)

#define GTK_GO_BOARD_IS_EMPTY(g) (!g)
#define GTK_GO_BOARD_IS_BLACK(g) (! (g & GTK_GO_BOARD_WHITE_MASK))
#define GTK_GO_BOARD_IS_WHITE(g) (g & GTK_GO_BOARD_WHITE_MASK)
#define GTK_GO_BOARD_IS_TRANS(g) (! (g & GTK_GO_BOARD_SOLID_MASK))
#define GTK_GO_BOARD_IS_SOLID(g) (g & GTK_GO_BOARD_SOLID_MASK)


/* `yes or no' markups.  */

#define GTK_GO_BOARD_MA_MASK (1 << 0) /* MA[] */
#define GTK_GO_BOARD_TR_MASK (1 << 1) /* TR[] */
#define GTK_GO_BOARD_CR_MASK (1 << 2) /* CR[] */
#define GTK_GO_BOARD_SQ_MASK (1 << 3) /* SQ[] */
#define GTK_GO_BOARD_SL_MASK (1 << 4) /* SL[] */
#define GTK_GO_BOARD_WHITE_SLASH_MASK (1 << 5) /* DD[] */
#define GTK_GO_BOARD_BLACK_SLASH_MASK (1 << 6) /* DD[] */

#define GTK_GO_BOARD_IS_MA(m) (m & GTK_GO_BOARD_MA_MASK)
#define GTK_GO_BOARD_IS_TR(m) (m & GTK_GO_BOARD_TR_MASK) 
#define GTK_GO_BOARD_IS_CR(m) (m & GTK_GO_BOARD_CR_MASK) 
#define GTK_GO_BOARD_IS_SQ(m) (m & GTK_GO_BOARD_SQ_MASK) 
#define GTK_GO_BOARD_IS_SL(m) (m & GTK_GO_BOARD_SL_MASK) 
#define GTK_GO_BOARD_IS_WHITE_SLASH(m) (m & GTK_GO_BOARD_WHITE_SLASH_MASK) 
#define GTK_GO_BOARD_IS_BLACK_SLASH(m) (m & GTK_GO_BOARD_BLACK_SLASH_MASK) 
#define GTK_GO_BOARD_IS_SLASH(m) (GTK_GO_BOARD_IS_WHITE_SLASH(m) || GTK_GO_BOARD_IS_BLACK_SLASH(m))

typedef enum _GtkGoBoardCoordStyle GtkGoBoardCoordStyle;
enum _GtkGoBoardCoordStyle
{
  GTK_GO_BOARD_NO_COORD,
  GTK_GO_BOARD_NORMAL_COORD,
  GTK_GO_BOARD_IGS_COORD
};

typedef struct _GtkGoBoardCell	GtkGoBoardCell;
struct _GtkGoBoardCell
{
  guint lineinfo;

  guint stone;
  guint markup;
  guint labelcolor;
  gchar *label;			/* LB[] */
  GSList *arrows;		/* AR[] */
  GSList *lines;		/* LN[] */
};


typedef struct _GtkGoBoard	GtkGoBoard;
struct _GtkGoBoard
{
  GtkDrawingArea drawing_area;

  guint high_contrast : 1;
  guint show_coordinates : 1;

  guint button_down : 1;

  GtkGoBoardCoordStyle coord_style;
  
  gint x_dim, y_dim;		/* FF[4] allow any rectangular board.  */

  gint press_x, press_y;

  /* drawing informations.  */
  gint cell_size;
  GdkRectangle board_rect;
  GdkRectangle upper_coordinates_rect;
  GdkRectangle lower_coordinates_rect;
  GdkRectangle left_coordinates_rect;
  GdkRectangle right_coordinates_rect;

  /* cell states. */
  GtkGoBoardCell *cells;
  gint pos_x, pos_y;

  /* images */
  GdkBitmap *solid_stone_mask;
  GdkBitmap *trans_stone_mask;
  GdkPixmap *black_stone_pixmap;
  GdkPixmap *white_stone_pixmap;

  GdkColor *grey_colors;
};

typedef struct _GtkGoBoardClass	GtkGoBoardClass;
struct _GtkGoBoardClass
{
  GtkDrawingAreaClass parent_class;

  void (* pressed)	(GtkGoBoard *go_board, gint x, gint y);
  void (* released)	(GtkGoBoard *go_board, gint x, gint y);
  void (* clicked)	(GtkGoBoard *go_board, gint x, gint y);
  void (* enter)	(GtkGoBoard *go_board, gint x, gint y);
  void (* leave)	(GtkGoBoard *go_board, gint x, gint y);
};


GtkType		gtk_go_board_get_type        (void);
GtkWidget*	gtk_go_board_new             (gint x_dim, gint y_dim);

void		gtk_go_board_set_high_contrast    (GtkGoBoard *go_board,
						   gint high_contrast);
void		gtk_go_board_set_show_coordinates (GtkGoBoard *go_board,
						   gint show_coordinates);

void		gtk_go_board_set_coord_style (GtkGoBoard *go_board,
					      GtkGoBoardCoordStyle style);

/* APIs...  */

void		gtk_go_board_set_stone	(GtkGoBoard *go_board,
					 gint x,
					 gint y,
					 guint stone);
guint		gtk_go_board_get_stone  (GtkGoBoard *go_board,
					 gint x,
					 gint y);

void		gtk_go_board_set_markup (GtkGoBoard *go_board,
					 gint x,
					 gint y,
					 guint markup);
guint		gtk_go_board_get_markup (GtkGoBoard *go_board,
					 gint x,
					 gint y);
void		gtk_go_board_add_markup (GtkGoBoard *go_board,
					 gint x,
					 gint y,
					 guint markup);
void		gtk_go_board_delete_markup (GtkGoBoard *go_board,
					    gint x,
					    gint y,
					    guint markup);
void		gtk_go_board_set_mark (GtkGoBoard *go_board,
			    gint        x,
			    gint        y,
			    guint       markup);

void		gtk_go_board_set_label (GtkGoBoard *go_board,
					gint x,
					gint y,
					const gchar *label);
void		gtk_go_board_set_plabel (GtkGoBoard *go_board,
					gint x,
					gint y,
					gint point);
gchar*		gtk_go_board_get_label (GtkGoBoard *go_board,
					gint x,
					gint y);

gint		gtk_go_board_add_line (GtkGoBoard *go_board,
				       gint sx,
				       gint sy,
				       gint ex,
				       gint ey);
void		gtk_go_board_delete_line (GtkGoBoard *go_board,
					  gint tag);

gint		gtk_go_board_add_arrow (GtkGoBoard *go_board,
					gint sx,
					gint sy,
					gint ex,
					gint ey);
void		gtk_go_board_delete_arrow (GtkGoBoard *go_board,
					   gint tag);

gint		gtk_go_board_add_line  (GtkGoBoard *go_board,
					gint sx,
					gint sy,
					gint ex,
					gint ey);
void		gtk_go_board_delete_line  (GtkGoBoard *go_board,
					   gint tag);

void	gtk_go_board_clear_board (GtkGoBoard *go_board);


/* Signal emits... */

void		gtk_go_board_pressed (GtkGoBoard *go_board,
				      gint x,
				      gint y, gint bt);
void		gtk_go_board_released (GtkGoBoard *go_board,
				       gint x,
				       gint y, gint bt);
void		gtk_go_board_clicked (GtkGoBoard *go_board,
				      gint x,
				      gint y);

void		gtk_go_board_enter   (GtkGoBoard *go_board,
				      gint x,
				      gint y);
void		gtk_go_board_leave   (GtkGoBoard *go_board,
				      gint x,
				      gint y);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* __GTK_GO_BOARD_H__ */

