/* Baduki - Computer Baduk(Go) Program 
   Copyright(c) 1998-1999 Lim Jaebum <artist@soback.kornet21.net>

   This program is free software; you can redistribute it and'or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
   USA */

/* $Id: mkpat.c,v 1.6 1999/04/01 20:24:58 artist Exp artist $ */ 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifndef TRUE
#define TRUE	1
#endif

#ifndef FALSE
#define FALSE	0
#endif

FILE *fdat, *fc, *fh;

char *CODE_HEADER =
" /* This file is automatially generated by mkpat from %s */\n"
" /* Do not edit this file manually */\n"
"\n"
"#include \"baduki.h\"\n"
"#include \"patterns.h\"\n"
"#include \"patternsdat.h\"\n"
"#include <stdlib.h>\n"
"\n"
"extern stringRec sList[maxString];\n"
"extern int SMap[maxString];\n"
"\n"
"extern int board_size;\n"
"\n"
"float pattern_value(int x, int y, struct pattern_type pat, int dir, int *px, int *py)\n"
"{\n"
"  float value = pat.value;\n"
"  int xx, yy;\n"
"\n"
"  switch (pat.id) {\n";

char *CODE_FOOTER =
"  default:\n"
"    break;\n"
"  }\n"
"\n"
"  restoreState();\n"
"  return value;\n"
"}\n";

char *PAT_HEADER =
" /* This file is automatially generated by mkpat from %s */\n"
" /* Do not edit this file manually */\n"
"\n"
"struct pattern_type pattern[] = {\n";

char *PAT_FOOTER =
"};\n";

#define PMAX	10

int id, xpos, ypos, xsize, ysize, pline, line = 0, pstart = -1;
float value;
char pattern[PMAX][PMAX];
char ptype[256];
int maxsize = 0;

void save_pat(int last_data)
{
  int i, j;
  char *dx;
  
  if (id == -1) {
    fprintf(stderr, "Patterns has no ID in line %d\n", line);
    exit(9);
  }
  
  ysize = pline;
  for (i = 0; i < ysize; i++) {
    if (xsize != strlen(pattern[i])) {
      fprintf(stderr, "Invallid pattern #%d in line %d\n", id, pstart+i);
      exit(9);
    }
    if ((dx = strchr(pattern[i], '@'))) {
      xpos = (dx - pattern[i]);
      ypos = ysize - i - 1;
      pattern[i][xpos] = '.';
    }
    for (j = 0; pattern[i][j]; j++)
      if (!strchr(".oOxXsStT%*", pattern[i][j])) {
        fprintf(stderr, "Invallid pattern #%d in line %d\n", id, pstart+i);
        exit(9);
      }
  }
  
  if (value > 0 && (xpos == -1 || ypos == -1)) {
    fprintf(stderr, "No Postion #%d in line %d\n", id, line);
    exit(9);
  }
  
  if (xsize * ysize > maxsize)
    maxsize = xsize * ysize;
  
  if (xsize < 2 || ysize < 2) {
    fprintf(stderr, "Pattern is too small #%d in line %d\n", id, line);
    exit(9);
  }
  
  fprintf(fh, "  {%d, %s, %d, %d, %d, %d, %f,\n", id,
  	ptype[0] ? ptype : "PT_CENTER",
  	xsize, ysize, xpos, ypos, value);
  
  for (i = ysize - 1; i >= 0; i--)
    fprintf(fh, "	\"%s\"\n", pattern[i]);
  
  if (last_data)
    fprintf(fh, "  }\n");
  else
    fprintf(fh, "  },\n");
  
  id = -1;
  ptype[0] = '\0';
  pstart = -1;
}

void strip_line(char *buf)
{
  int i;
  
  for (i = strlen(buf)-1; i >= 0; i--)
    if (buf[i] == ' ' || buf[i] == '\r' || buf[i] == '\n' || buf[i] == '\t')
      buf[i] = '\0';
    else
      return;
}
          
int main(int argc, char **argv)
{
  char fname[256], buf[1024];
  int in_pat=FALSE, in_code=FALSE, parg, patterns = 0, done = FALSE;

  if (argc != 4) {
    printf("Usage: %s data_file function_file header_file\n",
    	argv[0]);
    exit(1);
  }
  
  if (!(fdat = fopen(argv[1], "r"))) {
    perror("No datafile");
    exit(2);
  }
  
  if (!(fc = fopen(argv[2], "wt"))) {
    perror("in creating function_file");
    exit(3);
  }
  
  fprintf(fc, CODE_HEADER, argv[1]);

  if (!(fh = fopen(argv[3], "wt"))) {
    perror("in creating function_file");
    exit(3);
  }
  
  fprintf(fh, PAT_HEADER, argv[1]);
  
  done = FALSE;
  
  while(!done) {
    if (!fgets(buf, 1024, fdat)) {
      fprintf(stderr, "Unexpected end of file\n");
      exit(1);
    }
    line++;
    
    strip_line(buf);

    if (!*buf || *buf == ';')
      continue;
    
    if (in_code) {
      if (*buf == '#') {
        fprintf(fc, "    break;\n");
        in_code = FALSE;
      } else {
        fprintf(fc, "%s\n", buf);
        continue;
      }
    }
    
    switch(*buf) {
    case ';':
      continue;
      break;
    case '#':
      if (buf[1] == '.') {
        if (in_pat)
          save_pat(TRUE);
        done = TRUE;
        break;
      } else {
        patterns++;
        if (in_pat)
          save_pat(FALSE);
      }
        
      in_pat = TRUE;
      in_code = FALSE;
      pline = 0;
      
      parg = sscanf(buf, "#%d %f %d %d", &id, &value, &xpos, &ypos);
      if (parg == 2) {
        xpos = -1;
        ypos = -1;
      } else if (parg != 4) {
        fprintf(stderr, "Parse error in line %d\n", line);
      }
      
      break;
    case '>':
      if (!in_pat) {
        fprintf(stderr, "code without pattern in line %d\n", line);
        exit(9);
      }
      if (id < 0) {
        fprintf(stderr, "code without ID in line %d\n", line);
        exit(9);
      }
      in_code = TRUE;
      fprintf(fc, "  case %d:\n", id);
      fprintf(fc, "%s\n", buf+1);
      break;
    case '|':
      sprintf(ptype, buf+1);
      break;
    default:	/* Pattern */
      if (pstart == -1)
        pstart = line;
      xsize = strlen(buf);
      if (xsize > PMAX) {
        fprintf(stderr, "Pattern is to big #%d in line %d\n", id, line);
        exit(9);
      }
      sprintf(pattern[pline], buf);
      pline++;
      if (pline > PMAX) {
        fprintf(stderr, "Pattern is to big #%d in line %d\n", id, line);
        exit(9);
      }
      break;
    }
    
  }
  
  fprintf(fc, CODE_FOOTER);
  fprintf(fh, PAT_FOOTER);
  
  fprintf(fc, "int total_patterns = %d;\n", patterns);

  fclose(fdat);
  fclose(fc);
  fclose(fh);
  
  printf("Total Patterns: %d\n", patterns);
  printf("Max Size: %d\n", maxsize);

  return 0;
}
